@@ -245,26 +245,18 @@ vinagre_cmd_view_scaling (GtkAction     *action,
 			  VinagreWindow *window)
 {
   gboolean active;
+  VinagreTab *tab;
 
   g_return_if_fail (VINAGRE_IS_WINDOW (window));
 
+  tab = vinagre_window_get_active_tab (window);
+  if (!tab)
+    return;
+
   active = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
 
-  if (active &&
-      gdk_screen_is_composited (gtk_widget_get_screen (GTK_WIDGET (window))))
-    {
-      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
-      vinagre_utils_show_error (_("Scaling does not work properly on composited windows. Disable the visual effects and try again."),
-				GTK_WINDOW (window));
-      return;
-    }
-
-  if (!vinagre_tab_set_scaling (vinagre_window_get_active_tab (window), active))
-    {
-      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
-      vinagre_utils_show_error (_("Scaling is not supported on this installation.\n\nRead the README file (shipped with Vinagre) in order to know how to enable this feature."),
-				GTK_WINDOW (window));
-    }
+  if (!vinagre_tab_set_scaling (tab, active))
+    gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
 }
 
 void
@@ -291,6 +283,7 @@ vinagre_cmd_open_bookmark (VinagreWindow     *window,
   VinagreConnection *new_conn;
 
   g_return_if_fail (VINAGRE_IS_WINDOW (window));
+  g_return_if_fail (VINAGRE_IS_CONNECTION (conn));
 
   tab = vinagre_window_conn_exists (window, conn);
   if (tab)
@@ -375,138 +368,20 @@ vinagre_cmd_bookmarks_open (GtkAction     *action,
   vinagre_cmd_open_bookmark (window, conn);
 }
 
-/* Make url in about dialog clickable */
-static void
-vinagre_about_dialog_handle_url (GtkAboutDialog *about,
-				 const char     *link,
-				 gpointer        data)
-{
-  GError *error = NULL;
-  gchar  *address, *command;
-
-  VinagreWindow *window = VINAGRE_WINDOW (data);
-
-  if (g_strstr_len (link, strlen (link), "@"))
-    address = g_strdup_printf ("mailto:%s", link);
-  else
-    address = g_strdup (link);
-
-  command = g_strconcat ("gnome-open ", address,  NULL);
-	
-  gdk_spawn_command_line_on_screen (gtk_window_get_screen (GTK_WINDOW (window)),
-				    command,
-				    &error);
-  if (error != NULL) 
-    {
-      vinagre_utils_show_error (error->message, GTK_WINDOW (window));
-      g_error_free (error);
-    }
-
-  g_free (command);
-  g_free (address);
-}
-
 /* Help Menu */
 
 void
 vinagre_cmd_help_contents (GtkAction     *action,
 			   VinagreWindow *window)
 {
-  GError *error = NULL;
-  char *command;
-  const char *lang;
-  char *uri = NULL;
-  int i;
-
-  const char * const * langs = g_get_language_names ();
-
-  for (i = 0; langs[i]; i++)
-    {
-      lang = langs[i];
-      if (strchr (lang, '.')) 
-          continue;
-
-      if (uri)
-	g_free (uri);
-
-      uri = g_build_filename (DATADIR, "/gnome/help/vinagre/", lang, "/vinagre.xml", NULL);
-					
-      if (g_file_test (uri, G_FILE_TEST_EXISTS))
-          break;
-    }
-	
-  command = g_strconcat ("gnome-open ghelp://", uri,  NULL);
-	
-  gdk_spawn_command_line_on_screen (gtk_window_get_screen (GTK_WINDOW (window)),
-				    command,
-				    &error);
-  if (error != NULL) 
-    {
-      vinagre_utils_show_error (error->message, GTK_WINDOW (window));
-      g_error_free (error);
-    }
-
-  g_free (command);
-  g_free (uri);
+  vinagre_utils_help_contents (GTK_WINDOW (window));
 }
 
 void
 vinagre_cmd_help_about (GtkAction     *action,
 			VinagreWindow *window)
 {
-  static const gchar * const authors[] = {
-	"Jonh Wendell <jwendell@gnome.org>",
-	NULL
-  };
-
-  static const gchar * const artists[] = {
-	"Vinicius Depizzol <vdepizzol@gmail.com>",
-	NULL
-  };
-
-  static const gchar copyright[] = \
-	"Copyright \xc2\xa9 2007 Jonh Wendell";
-
-  static const gchar comments[] = \
-	N_("Vinagre is a VNC client for the GNOME Desktop");
-
-  static const char *license[] = {
-	N_("Vinagre is free software; you can redistribute it and/or modify "
-	   "it under the terms of the GNU General Public License as published by "
-	   "the Free Software Foundation; either version 2 of the License, or "
-	   "(at your option) any later version."),
-	N_("Vinagre is distributed in the hope that it will be useful, "
-	   "but WITHOUT ANY WARRANTY; without even the implied warranty of "
-	   "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the "
-	   "GNU General Public License for more details."),
-	N_("You should have received a copy of the GNU General Public License "
-	   "along with this program. If not, see <http://www.gnu.org/licenses/>.")
-  };
-
-  gchar *license_trans;
-
-  license_trans = g_strjoin ("\n\n", _(license[0]), _(license[1]),
-				     _(license[2]), NULL);
-
-  /* Make URLs and email clickable in about dialog */
-  gtk_about_dialog_set_url_hook (vinagre_about_dialog_handle_url, window, NULL);
-  gtk_about_dialog_set_email_hook (vinagre_about_dialog_handle_url, window, NULL);
-
-
-  gtk_show_about_dialog (GTK_WINDOW (window),
-			 "authors", authors,
-			 "artists", artists,
-			 "comments", _(comments),
-			 "copyright", copyright,
-			 "license", license_trans,
-			 "wrap-license", TRUE,
-			 "logo-icon-name", "vinagre",
-			 "translator-credits", _("translator-credits"),
-			 "version", VERSION,
-			 "website", "http://www.gnome.org/projects/vinagre/",
-			 "website-label", _("Vinagre Website"),
-			 NULL);
-  g_free (license_trans);
+  vinagre_utils_help_about (GTK_WINDOW (window));
 }
 
-/* vim: ts=8 */
+/* vim: set ts=8: */
