@@ -2,7 +2,7 @@
  * vinagre-tab.c
  * This file is part of vinagre
  *
- * Copyright (C) 2007 - Jonh Wendell <wendell@bani.com.br>
+ * Copyright (C) 2007,2008 - Jonh Wendell <wendell@bani.com.br>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,6 +22,7 @@
 #include <config.h>
 #endif
 
+#include <gdk/gdkkeysyms.h>
 #include <glib/gi18n.h>
 #include <glade/glade.h>
 #include <gnome-keyring.h>
@@ -30,6 +31,8 @@
 #include "vinagre-notebook.h"
 #include "vinagre-tab.h"
 #include "vinagre-utils.h"
+#include "vinagre-prefs.h"
+#include "view/autoDrawer.h"
 
 #define VINAGRE_TAB_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), VINAGRE_TYPE_TAB, VinagreTabPrivate))
 
@@ -40,10 +43,15 @@ struct _VinagreTabPrivate
   VinagreConnection *conn;
   VinagreNotebook   *nb;
   VinagreWindow     *window;
-  GtkStatusbar      *status;
-  guint              status_id;
   gboolean           save_password;
   guint32            keyring_item_id;
+  VinagreTabState    state;
+  gchar             *clipboard_str;
+  GtkWidget         *layout;
+  GtkWidget         *toolbar;
+  GtkWidget         *ro_button;
+  GtkWidget         *scaling_button;
+  gboolean           pointer_grab;
 };
 
 G_DEFINE_TYPE(VinagreTab, vinagre_tab, GTK_TYPE_VBOX)
@@ -98,16 +106,19 @@ vinagre_tab_window_state_cb (GtkWidget           *widget,
 
   if (event->new_window_state & GDK_WINDOW_STATE_FULLSCREEN)
     {
-      vnc_display_set_pointer_grab  (VNC_DISPLAY(tab->priv->vnc), TRUE);
+      gtk_widget_show (tab->priv->toolbar);
+      ViewAutoDrawer_SetActive (VIEW_AUTODRAWER (tab->priv->layout), TRUE);
       gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (tab->priv->scroll),
 				      h, v);
     }
   else
     {
-      vnc_display_set_pointer_grab  (VNC_DISPLAY(tab->priv->vnc), FALSE);
+      gtk_widget_hide (tab->priv->toolbar);
+      ViewAutoDrawer_SetActive (VIEW_AUTODRAWER (tab->priv->layout), FALSE);
       gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (tab->priv->scroll),
 				      GTK_POLICY_AUTOMATIC,
 				      GTK_POLICY_AUTOMATIC);
+      vinagre_notebook_show_hide_tabs (tab->priv->nb);
     }
 
   return FALSE;
@@ -124,10 +135,10 @@ vinagre_tab_get_property (GObject    *object,
   switch (prop_id)
     {
       case PROP_CONN:
-        g_value_set_pointer (value, tab->priv->conn);
+        g_value_set_object (value, tab->priv->conn);
 	break;
       case PROP_WINDOW:
-        g_value_set_pointer (value, tab->priv->window);
+        g_value_set_object (value, tab->priv->window);
 	break;
       default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -146,10 +157,11 @@ vinagre_tab_set_property (GObject      *object,
   switch (prop_id)
     {
       case PROP_CONN:
-        tab->priv->conn = g_value_get_pointer (value);
+        tab->priv->conn = g_value_get_object (value);
+	g_object_set_data (G_OBJECT (tab->priv->conn), VINAGRE_TAB_KEY, tab);
         break;
       case PROP_WINDOW:
-        tab->priv->window = g_value_get_pointer (value);
+        tab->priv->window = g_value_get_object (value);
 	g_signal_connect (tab->priv->window,
 			  "window-state-event",
 			  G_CALLBACK (vinagre_tab_window_state_cb),
@@ -166,15 +178,15 @@ vinagre_tab_finalize (GObject *object)
 {
   VinagreTab *tab = VINAGRE_TAB (object);
 
-  gtk_statusbar_pop (tab->priv->status, tab->priv->status_id);
-  gtk_action_group_set_sensitive (vinagre_window_get_sensitive_action (tab->priv->window),
-				  TRUE);
-
   g_signal_handlers_disconnect_by_func (tab->priv->window,
   					vinagre_tab_window_state_cb,
   					tab);
-  vinagre_connection_free (tab->priv->conn);
-	
+  g_object_unref (tab->priv->conn);
+  tab->priv->conn = NULL;
+
+  g_free (tab->priv->clipboard_str);
+  tab->priv->clipboard_str = NULL;
+
   G_OBJECT_CLASS (vinagre_tab_parent_class)->finalize (object);
 }
 
@@ -189,24 +201,26 @@ vinagre_tab_class_init (VinagreTabClass *klass)
 
   g_object_class_install_property (object_class,
 				   PROP_CONN,
-				   g_param_spec_pointer ("conn",
-							 "Connection",
-							 "The connection",
-							 G_PARAM_READWRITE |
-							 G_PARAM_CONSTRUCT_ONLY |
-							 G_PARAM_STATIC_NAME |
-							 G_PARAM_STATIC_NICK |
-							 G_PARAM_STATIC_BLURB));
+				   g_param_spec_object ("conn",
+							"Connection",
+							"The connection",
+							VINAGRE_TYPE_CONNECTION,
+							G_PARAM_READWRITE |
+							G_PARAM_CONSTRUCT_ONLY |
+							G_PARAM_STATIC_NAME |
+							G_PARAM_STATIC_NICK |
+							G_PARAM_STATIC_BLURB));
   g_object_class_install_property (object_class,
 				   PROP_WINDOW,
-				   g_param_spec_pointer ("window",
-							 "Window",
-							 "The VinagreWindow",
-							 G_PARAM_READWRITE |
-							 G_PARAM_CONSTRUCT_ONLY |
-							 G_PARAM_STATIC_NAME |
-							 G_PARAM_STATIC_NICK |
-							 G_PARAM_STATIC_BLURB));
+				   g_param_spec_object ("window",
+							"Window",
+							"The VinagreWindow",
+							VINAGRE_TYPE_WINDOW,
+							G_PARAM_READWRITE |
+							G_PARAM_CONSTRUCT_ONLY |
+							G_PARAM_STATIC_NAME |
+							G_PARAM_STATIC_NICK |
+							G_PARAM_STATIC_BLURB));
 
   signals[TAB_CONNECTED] =
 		g_signal_new ("tab-connected",
@@ -241,24 +255,22 @@ vinagre_tab_class_init (VinagreTabClass *klass)
   g_type_class_add_private (object_class, sizeof (VinagreTabPrivate));
 }
 
-static gboolean
+static void
 open_vnc (VinagreTab *tab)
 {
   gchar *port;
 
-  tab->priv->status = GTK_STATUSBAR (vinagre_window_get_statusbar (tab->priv->window));
-  tab->priv->status_id = gtk_statusbar_get_context_id (tab->priv->status, "VNC Widget Messages");
-  
-  port = g_strdup_printf ("%d", tab->priv->conn->port);
-  
-  if (!vnc_display_open_host (VNC_DISPLAY(tab->priv->vnc), tab->priv->conn->host, port))
+  vnc_display_set_force_size (VNC_DISPLAY(tab->priv->vnc),
+			      !vinagre_connection_get_scaling (tab->priv->conn));
+
+  port = g_strdup_printf ("%d", vinagre_connection_get_port (tab->priv->conn));
+
+  if (vnc_display_open_host (VNC_DISPLAY(tab->priv->vnc), vinagre_connection_get_host (tab->priv->conn), port))
+    gtk_widget_grab_focus (tab->priv->vnc);
+  else
     vinagre_utils_show_error (_("Error connecting to host."), NULL);
 
-  vnc_display_set_pointer_local (VNC_DISPLAY(tab->priv->vnc), TRUE);
-
   g_free (port);
-  gtk_widget_grab_focus (tab->priv->vnc);
-  return FALSE;
 }
 
 static void
@@ -273,8 +285,6 @@ vnc_connected_cb (VncDisplay *vnc, VinagreTab *tab)
 static void
 vnc_disconnected_cb (VncDisplay *vnc, VinagreTab *tab)
 {
-  _vinagre_window_del_machine_connected (tab->priv->window);
-
   /* Emits the signal saying that we have disconnected from the machine */
   g_signal_emit (G_OBJECT (tab),
 		 signals[TAB_DISCONNECTED],
@@ -288,7 +298,7 @@ vnc_auth_failed_cb (VncDisplay *vnc, const gchar *msg, VinagreTab *tab)
   gchar   *name;
 
   message = g_string_new (NULL);
-  name = vinagre_connection_best_name (vinagre_tab_get_conn (tab));
+  name = vinagre_connection_get_best_name (vinagre_tab_get_conn (tab));
 
   g_string_printf (message, _("Authentication to host \"%s\" has failed"),
 		   name);
@@ -316,7 +326,7 @@ vnc_auth_unsupported_cb (VncDisplay *vnc, guint auth_type, VinagreTab *tab)
   gchar   *name;
 
   message = g_string_new (NULL);
-  name = vinagre_connection_best_name (vinagre_tab_get_conn (tab));
+  name = vinagre_connection_get_best_name (vinagre_tab_get_conn (tab));
 
   g_string_printf (message, _("Authentication method to host \"%s\" is unsupported. (%u)"),
 		   name,
@@ -329,22 +339,45 @@ vnc_auth_unsupported_cb (VncDisplay *vnc, guint auth_type, VinagreTab *tab)
   vinagre_notebook_remove_tab (tab->priv->nb, tab);
 }
 
+/* text was actually requested */
+static void
+copy_cb (GtkClipboard     *clipboard,
+         GtkSelectionData *data,
+	 guint             info,
+	 VinagreTab       *tab)
+{
+  gtk_selection_data_set_text (data, tab->priv->clipboard_str, -1);
+}
+
 static void
 vnc_server_cut_text_cb (VncDisplay *vnc, const gchar *text, VinagreTab *tab)
 {
   GtkClipboard *cb;
-  gchar *out;
   gsize a, b;
+  GtkTargetEntry targets[] = {
+				{"UTF8_STRING", 0, 0},
+				{"COMPOUND_TEXT", 0, 0},
+				{"TEXT", 0, 0},
+				{"STRING", 0, 0},
+			     };
 
   if (!text)
     return;
 
-  out = g_convert (text, -1, "utf-8", "iso8859-1", &a, &b, NULL);
-  if (out) {
-    cb = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
-    gtk_clipboard_set_text (cb, out, -1);
-    g_free (out);
-  }
+  g_free (tab->priv->clipboard_str);
+  tab->priv->clipboard_str = g_convert (text, -1, "utf-8", "iso8859-1", &a, &b, NULL);
+
+  if (tab->priv->clipboard_str)
+    {
+      cb = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
+
+      gtk_clipboard_set_with_owner (cb,
+				    targets,
+				    G_N_ELEMENTS(targets),
+				    (GtkClipboardGetFunc) copy_cb,
+				    NULL,
+				    G_OBJECT (tab));
+    }
 }
 
 static void
@@ -363,9 +396,11 @@ vinagre_tab_add_recent_used (VinagreTab *tab)
   data = g_slice_new (GtkRecentData);
 
   uri = g_string_new (NULL);
-  g_string_printf (uri, "%s:%d", tab->priv->conn->host, tab->priv->conn->port);
+  g_string_printf (uri, "vnc://%s:%d",
+                   vinagre_connection_get_host (tab->priv->conn),
+                   vinagre_connection_get_port (tab->priv->conn));
 
-  data->display_name = vinagre_connection_best_name (tab->priv->conn);
+  data->display_name = vinagre_connection_get_best_name (tab->priv->conn);
   data->description = NULL;
   data->mime_type = g_strdup ("application/x-remote-connection");
   data->app_name = (gchar *) g_get_application_name ();
@@ -396,12 +431,12 @@ vinagre_tab_save_password (VinagreTab *tab)
                 NULL,           /* default keyring */
                 NULL,           /* user            */
                 NULL,           /* domain          */
-                tab->priv->conn->host,   /* server          */
+                vinagre_connection_get_host (tab->priv->conn),   /* server          */
                 NULL,           /* object          */
                 "rfb",          /* protocol        */
                 "vnc-password", /* authtype        */
-                tab->priv->conn->port,           /* port            */
-                tab->priv->conn->password,       /* password        */
+                vinagre_connection_get_port (tab->priv->conn),           /* port            */
+                vinagre_connection_get_password (tab->priv->conn),       /* password        */
                 &tab->priv->keyring_item_id);
 
   if (result != GNOME_KEYRING_RESULT_OK)
@@ -424,11 +459,11 @@ vinagre_tab_find_password (VinagreTab *tab)
   result = gnome_keyring_find_network_password_sync (
                 NULL,           /* user     */
 		NULL,           /* domain   */
-		tab->priv->conn->host,   /* server   */
+		vinagre_connection_get_host (tab->priv->conn),   /* server   */
 		NULL,           /* object   */
 		"rfb",          /* protocol */
 		"vnc-password", /* authtype */
-		tab->priv->conn->port,           /* port     */
+		vinagre_connection_get_port (tab->priv->conn),           /* port     */
 		&matches);
 
   if (result != GNOME_KEYRING_RESULT_OK || matches == NULL || matches->data == NULL)
@@ -449,11 +484,27 @@ vnc_initialized_cb (VncDisplay *vnc, VinagreTab *tab)
 {
   GtkLabel *label;
   gchar    *name;
+  gboolean scaling, view_only, fullscreen;
+
+  g_object_get (tab->priv->conn,
+		"view-only", &view_only,
+		"scaling", &scaling,
+		"fullscreen", &fullscreen,
+		NULL);
+
+  vinagre_tab_set_scaling (tab, scaling);
+  vinagre_tab_set_readonly (tab, view_only);
+  vnc_display_set_pointer_local (VNC_DISPLAY(tab->priv->vnc), TRUE);
+  vnc_display_set_keyboard_grab (VNC_DISPLAY(tab->priv->vnc), TRUE);
+  vnc_display_set_pointer_grab (VNC_DISPLAY(tab->priv->vnc), TRUE);
+
+  if (fullscreen)
+    vinagre_window_toggle_fullscreen (tab->priv->window);
 
   vinagre_connection_set_desktop_name (tab->priv->conn,
 				       vnc_display_get_name (VNC_DISPLAY (tab->priv->vnc)));
 
-  name = vinagre_connection_best_name (tab->priv->conn);
+  name = vinagre_connection_get_best_name (tab->priv->conn);
   label = g_object_get_data (G_OBJECT (tab), "label");
   g_return_if_fail (label != NULL);
   gtk_label_set_label (label, name);
@@ -463,11 +514,8 @@ vnc_initialized_cb (VncDisplay *vnc, VinagreTab *tab)
   vinagre_tab_save_password (tab);
   vinagre_tab_add_recent_used (tab);
 
-  gtk_statusbar_push (tab->priv->status,
-		      tab->priv->status_id,
-		      _("Press Ctrl+Alt to grab the cursor"));
-
-  _vinagre_window_add_machine_connected (tab->priv->window);
+  tab->priv->state = VINAGRE_TAB_STATE_CONNECTED;
+  vinagre_window_update_machine_menu_sensitivity (tab->priv->window);
 
   /* Emits the signal saying that we have connected to the machine */
   g_signal_emit (G_OBJECT (tab),
@@ -491,7 +539,7 @@ ask_password(VinagreTab *tab)
   gtk_window_set_transient_for (GTK_WINDOW(password_dialog), GTK_WINDOW(tab->priv->window));
 
   host_label = glade_xml_get_widget (xml, "host_label");
-  name = vinagre_connection_best_name (tab->priv->conn);
+  name = vinagre_connection_get_best_name (tab->priv->conn);
   gtk_label_set_text (GTK_LABEL (host_label), name);
   g_free (name);
 
@@ -515,40 +563,50 @@ static void
 vnc_authentication_cb (VncDisplay *vnc, GValueArray *credList, VinagreTab *tab)
 {
   gchar *password;
+  int i;
 
-  password = vinagre_tab_find_password (tab);
-  if (!password)
-    {
-      password = ask_password (tab);
-      if (!password) {
-        vinagre_notebook_remove_tab (tab->priv->nb, tab);
-        return;
+  if (credList == NULL)
+    return;
+
+  for (i = 0; i < credList->n_values; i++) {
+    switch (g_value_get_enum (&credList->values[i]))
+      {
+        case VNC_DISPLAY_CREDENTIAL_PASSWORD:
+          password = vinagre_tab_find_password (tab);
+          if (!password)
+            {
+              password = ask_password (tab);
+              if (!password) {
+                vinagre_notebook_remove_tab (tab->priv->nb, tab);
+                return;
+              }
+            }
+
+          vinagre_connection_set_password (tab->priv->conn, password);
+          vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_PASSWORD, password);
+
+          g_free (password);
+          break;
+
+        case VNC_DISPLAY_CREDENTIAL_CLIENTNAME:
+          vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_CLIENTNAME, "vinagre");
+          break;
       }
-    }
-
-  vinagre_connection_set_password (tab->priv->conn, password);
-  vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_PASSWORD, password);
-
-  g_free (password);
+  }
 }
 
-static void vnc_grab_cb (VncDisplay *vnc, VinagreTab *tab)
+static void
+vnc_pointer_grab_cb (VncDisplay *vnc, VinagreTab *tab)
 {
+  tab->priv->pointer_grab = TRUE;
+  vinagre_window_set_title (tab->priv->window);
+}
 
-  gtk_statusbar_pop (tab->priv->status,
-		     tab->priv->status_id);
-
-  gtk_statusbar_push (tab->priv->status,
-		      tab->priv->status_id,
-		      _("Press Ctrl+Alt to release the cursor"));
-
-  gtk_action_group_set_sensitive (vinagre_window_get_main_action (tab->priv->window),
-				  FALSE);
-  gtk_action_group_set_sensitive (vinagre_window_get_sensitive_action (tab->priv->window),
-				  FALSE);
-
-  if (vinagre_window_is_fullscreen (tab->priv->window))
-    gtk_notebook_set_show_tabs (GTK_NOTEBOOK (tab->priv->nb), FALSE);
+static void
+vnc_pointer_ungrab_cb (VncDisplay *vnc, VinagreTab *tab)
+{
+  tab->priv->pointer_grab = FALSE;
+  vinagre_window_set_title (tab->priv->window);
 }
 
 static void vnc_bell_cb (VncDisplay *vnc, VinagreTab *tab)
@@ -556,34 +614,149 @@ static void vnc_bell_cb (VncDisplay *vnc, VinagreTab *tab)
   gdk_window_beep (GTK_WIDGET (tab->priv->window)->window);
 }
 
-static void vnc_ungrab_cb (VncDisplay *vnc, VinagreTab *tab)
+static void
+close_button_clicked (GtkToolButton *button,
+		      VinagreTab    *tab)
 {
-  gtk_statusbar_pop (tab->priv->status, tab->priv->status_id);
-  gtk_statusbar_push (tab->priv->status,
-		      tab->priv->status_id,
-		      _("Press Ctrl+Alt to grab the cursor"));
+  vinagre_notebook_remove_tab (tab->priv->nb, tab);
+}
 
-  gtk_action_group_set_sensitive (vinagre_window_get_main_action (tab->priv->window),
-				  TRUE);
-  gtk_action_group_set_sensitive (vinagre_window_get_sensitive_action (tab->priv->window),
-				  TRUE);
+static void
+fullscreen_button_clicked (GtkToolButton *button,
+			   VinagreTab    *tab)
+{
+  vinagre_window_toggle_fullscreen (tab->priv->window);
+}
 
-  if (vinagre_window_is_fullscreen (tab->priv->window))
-    gtk_notebook_set_show_tabs (GTK_NOTEBOOK (tab->priv->nb), TRUE);
+static void
+screenshot_button_clicked (GtkToolButton *button,
+			   VinagreTab    *tab)
+{
+  vinagre_tab_take_screenshot (tab);
+}
+
+static void
+cad_button_clicked (GtkToolButton *button,
+			   VinagreTab    *tab)
+{
+  vinagre_tab_send_ctrlaltdel (tab);
+}
+
+static void
+ro_button_clicked (GtkToggleToolButton *button,
+		   VinagreTab          *tab)
+{
+  GtkAction *action;
+
+  vinagre_tab_set_readonly (tab, gtk_toggle_tool_button_get_active (button));
+
+  action = gtk_action_group_get_action (vinagre_window_get_connected_action (tab->priv->window), "ViewReadOnly");
+  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
+				vinagre_tab_get_readonly (tab));
+}
+
+static void
+scaling_button_clicked (GtkToggleToolButton *button,
+			VinagreTab          *tab)
+{
+  GtkAction *action;
+
+  if (!vinagre_tab_set_scaling (tab, gtk_toggle_tool_button_get_active (button)))
+    gtk_toggle_tool_button_set_active (button, FALSE);
+
+  action = gtk_action_group_get_action (vinagre_window_get_connected_action (tab->priv->window), "ViewScaling");
+  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
+				vinagre_tab_get_scaling (tab));
+}
+
+static void
+setup_layout (VinagreTab *tab)
+{
+  GtkWidget  *button;
+
+  tab->priv->toolbar = gtk_toolbar_new ();
+  GTK_WIDGET_SET_FLAGS (tab->priv->toolbar, GTK_NO_SHOW_ALL);
+
+  gtk_toolbar_set_tooltips (GTK_TOOLBAR (tab->priv->toolbar), TRUE);
+  gtk_toolbar_set_style (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOLBAR_BOTH_HORIZ);
+
+  /* Leave fullscreen */
+  button = GTK_WIDGET (gtk_tool_button_new_from_stock (GTK_STOCK_LEAVE_FULLSCREEN));
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_tool_item_set_is_important (GTK_TOOL_ITEM (button), TRUE);
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+  g_signal_connect (button, "clicked", G_CALLBACK (fullscreen_button_clicked), tab);
+
+  /* Close connection */
+  button = GTK_WIDGET (gtk_tool_button_new_from_stock (GTK_STOCK_CLOSE));
+  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Close connection"));
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+  g_signal_connect (button, "clicked", G_CALLBACK (close_button_clicked), tab);
+
+  /* Space */
+  button = GTK_WIDGET (gtk_separator_tool_item_new ());
+  gtk_separator_tool_item_set_draw (GTK_SEPARATOR_TOOL_ITEM (button), FALSE);
+  gtk_tool_item_set_expand (GTK_TOOL_ITEM (button), TRUE);
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+
+  /* Scaling */
+  button = GTK_WIDGET (gtk_toggle_tool_button_new ());
+  gtk_tool_button_set_label (GTK_TOOL_BUTTON (button), _("Scaling"));
+  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Scaling"));
+  gtk_tool_button_set_icon_name (GTK_TOOL_BUTTON (button), "zoom-fit-best");
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+  g_signal_connect (button, "toggled", G_CALLBACK (scaling_button_clicked), tab);
+  tab->priv->scaling_button = button;
+
+  /* Read only */
+  button = GTK_WIDGET (gtk_toggle_tool_button_new ());
+  gtk_tool_button_set_label (GTK_TOOL_BUTTON (button), _("Read only"));
+  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Read only"));
+  gtk_tool_button_set_icon_name (GTK_TOOL_BUTTON (button), "emblem-readonly");
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+  g_signal_connect (button, "toggled", G_CALLBACK (ro_button_clicked), tab);
+  tab->priv->ro_button = button;
+
+  /* Screenshot */
+  button = GTK_WIDGET (gtk_tool_button_new (NULL, NULL));
+  gtk_tool_button_set_icon_name (GTK_TOOL_BUTTON (button), "applets-screenshooter");
+  gtk_tool_button_set_label (GTK_TOOL_BUTTON (button), _("Take screenshot"));
+  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Take screenshot"));
+  g_signal_connect (button, "clicked", G_CALLBACK (screenshot_button_clicked), tab);
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+
+  /* Send Ctrl-alt-del */
+  button = GTK_WIDGET (gtk_tool_button_new (NULL, NULL));
+  gtk_tool_button_set_icon_name (GTK_TOOL_BUTTON (button), "preferences-desktop-keyboard-shortcuts");
+  gtk_tool_button_set_label (GTK_TOOL_BUTTON (button), _("Send Ctrl-Alt-Del"));
+  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Send Ctrl-Alt-Del"));
+  g_signal_connect (button, "clicked", G_CALLBACK (cad_button_clicked), tab);
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+
+  tab->priv->layout = ViewAutoDrawer_New ();
+  ViewAutoDrawer_SetActive (VIEW_AUTODRAWER (tab->priv->layout), FALSE);
+  ViewOvBox_SetOver (VIEW_OV_BOX (tab->priv->layout), tab->priv->toolbar);
+  ViewOvBox_SetUnder (VIEW_OV_BOX (tab->priv->layout), tab->priv->scroll);
 }
 
 static void
 vinagre_tab_init (VinagreTab *tab)
 {
-  GtkWidget *align;
-  GtkWidget *viewport;
+  GtkWidget     *viewport;
+  gboolean       shared;
 
   tab->priv = VINAGRE_TAB_GET_PRIVATE (tab);
   tab->priv->save_password = FALSE;
   tab->priv->keyring_item_id = 0;
-
-  /* Create the alignment */
-  align = gtk_alignment_new (0.5, 0.5, 0, 0);
+  tab->priv->state = VINAGRE_TAB_STATE_INITIALIZING;
+  tab->priv->clipboard_str = NULL;
+  tab->priv->pointer_grab = FALSE;
 
   /* Create the scrolled window */
   tab->priv->scroll = gtk_scrolled_window_new (NULL, NULL);
@@ -593,17 +766,20 @@ vinagre_tab_init (VinagreTab *tab)
   gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (tab->priv->scroll),
 				       GTK_SHADOW_NONE);
 
-  gtk_box_pack_end (GTK_BOX(tab), tab->priv->scroll, TRUE, TRUE, 0);
-
   /* Create the vnc widget */
   tab->priv->vnc = vnc_display_new ();
-  gtk_container_add (GTK_CONTAINER (align), tab->priv->vnc);
 
   gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (tab->priv->scroll),
-					 align);
+					 tab->priv->vnc);
   viewport = gtk_bin_get_child (GTK_BIN (tab->priv->scroll));
   gtk_viewport_set_shadow_type(GTK_VIEWPORT (viewport), GTK_SHADOW_NONE);
 
+  g_object_get (vinagre_prefs_get_default (),
+		"shared-flag", &shared,
+		NULL);
+  vnc_display_set_shared_flag (VNC_DISPLAY (tab->priv->vnc),
+			       shared);
+
   g_signal_connect (tab->priv->vnc,
 		    "vnc-connected",
 		    G_CALLBACK (vnc_connected_cb),
@@ -626,12 +802,12 @@ vinagre_tab_init (VinagreTab *tab)
 
   g_signal_connect (tab->priv->vnc,
 		    "vnc-pointer-grab",
-		    G_CALLBACK (vnc_grab_cb),
+		    G_CALLBACK (vnc_pointer_grab_cb),
 		    tab);
 
   g_signal_connect (tab->priv->vnc,
 		    "vnc-pointer-ungrab",
-		    G_CALLBACK (vnc_ungrab_cb),
+		    G_CALLBACK (vnc_pointer_ungrab_cb),
 		    tab);
 
   g_signal_connect (tab->priv->vnc,
@@ -654,12 +830,9 @@ vinagre_tab_init (VinagreTab *tab)
 		    G_CALLBACK (vnc_bell_cb),
 		    tab);
 
- /* connect VNC */
- /* FIXME: i had to add a timeout because private conn is not available at this time*/
-  g_timeout_add (1000,
-		 (GSourceFunc) open_vnc,
-		 tab);
+  setup_layout (tab);
 
+  gtk_box_pack_end (GTK_BOX(tab), tab->priv->layout, TRUE, TRUE, 0);
   gtk_widget_show_all (GTK_WIDGET (tab));
 }
 
@@ -670,6 +843,7 @@ vinagre_tab_new (VinagreConnection *conn, VinagreWindow *window)
 				   "conn", conn,
 				   "window", window,
 				   NULL);
+  open_vnc (tab);
   return GTK_WIDGET (tab);
 }
 
@@ -687,8 +861,8 @@ vinagre_tab_get_tooltips (VinagreTab *tab)
 				  "<b>%s</b> %d\n"
 				  "<b>%s</b> %dx%d",
 				  _("Desktop Name:"), vnc_display_get_name (VNC_DISPLAY (tab->priv->vnc)),
-				  _("Host:"), tab->priv->conn->host,
-				  _("Port:"), tab->priv->conn->port,
+				  _("Host:"), vinagre_connection_get_host (tab->priv->conn),
+				  _("Port:"), vinagre_connection_get_port (tab->priv->conn),
 				  _("Dimensions:"), vnc_display_get_width (VNC_DISPLAY (tab->priv->vnc)), vnc_display_get_height (VNC_DISPLAY (tab->priv->vnc)));
 
   return tip;
@@ -754,7 +928,7 @@ vinagre_tab_take_screenshot (VinagreTab *tab)
   pix = vnc_display_get_pixbuf (VNC_DISPLAY (tab->priv->vnc));
 
   filename = NULL;
-  name = vinagre_connection_best_name (tab->priv->conn);
+  name = vinagre_connection_get_best_name (tab->priv->conn);
   suggested_filename = g_string_new (NULL);
   g_string_printf (suggested_filename, _("Screenshot of %s"), name);
   g_string_append (suggested_filename, ".png");
@@ -787,6 +961,16 @@ vinagre_tab_take_screenshot (VinagreTab *tab)
   g_free (name);
 }
 
+void
+vinagre_tab_send_ctrlaltdel (VinagreTab *tab)
+{
+  guint keys[] = { GDK_Control_L, GDK_Alt_L, GDK_Delete };
+
+  g_return_if_fail (VINAGRE_IS_TAB (tab));
+
+  vnc_display_send_keys_ex (VNC_DISPLAY (tab->priv->vnc), keys, sizeof (keys) / sizeof (keys[0]), VNC_DISPLAY_KEY_EVENT_CLICK);
+}
+
 void
 vinagre_tab_paste_text (VinagreTab *tab, const gchar *text)
 {
@@ -803,4 +987,91 @@ vinagre_tab_paste_text (VinagreTab *tab, const gchar *text)
     }
 }
 
-/* vim: ts=8 */
+gboolean
+vinagre_tab_set_scaling (VinagreTab *tab, gboolean active) {
+  g_return_val_if_fail (VINAGRE_IS_TAB (tab), FALSE);
+
+  if (vnc_display_get_scaling (VNC_DISPLAY (tab->priv->vnc)) == active)
+    return TRUE;
+
+  if (active &&
+      gdk_screen_is_composited (gtk_widget_get_screen (GTK_WIDGET (tab->priv->window))))
+    {
+      vinagre_utils_show_error (_("Scaling does not work properly on composited windows. Disable the visual effects and try again."),
+				GTK_WINDOW (tab->priv->window));
+      return FALSE;
+    }
+
+  vnc_display_set_force_size (VNC_DISPLAY(tab->priv->vnc), !active);
+  if (!vnc_display_set_scaling (VNC_DISPLAY (tab->priv->vnc), active))
+    {
+      vinagre_utils_show_error (_("Scaling is not supported on this installation.\n\nRead the README file (shipped with Vinagre) in order to know how to enable this feature."),
+				GTK_WINDOW (tab->priv->window));
+      return FALSE;
+    }
+
+  gtk_toggle_tool_button_set_active (GTK_TOGGLE_TOOL_BUTTON (tab->priv->scaling_button),
+				     active);
+
+  if (active)
+    gtk_widget_set_size_request (tab->priv->vnc, 0, 0);
+  else
+    gtk_widget_set_size_request (tab->priv->vnc,
+				 vnc_display_get_width (VNC_DISPLAY (tab->priv->vnc)),
+				 vnc_display_get_height (VNC_DISPLAY (tab->priv->vnc)));
+
+  return TRUE;
+}
+
+gboolean
+vinagre_tab_get_scaling (VinagreTab *tab) {
+  g_return_if_fail (VINAGRE_IS_TAB (tab));
+
+  return vnc_display_get_scaling (VNC_DISPLAY (tab->priv->vnc));
+}
+
+void
+vinagre_tab_set_readonly (VinagreTab *tab, gboolean active) {
+  g_return_if_fail (VINAGRE_IS_TAB (tab));
+
+  vnc_display_set_read_only (VNC_DISPLAY (tab->priv->vnc), active);
+  gtk_toggle_tool_button_set_active (GTK_TOGGLE_TOOL_BUTTON (tab->priv->ro_button),
+				     active);
+
+}
+
+gboolean
+vinagre_tab_get_readonly (VinagreTab *tab) {
+  g_return_if_fail (VINAGRE_IS_TAB (tab));
+
+  return vnc_display_get_read_only (VNC_DISPLAY (tab->priv->vnc));
+}
+
+VinagreTabState
+vinagre_tab_get_state (VinagreTab *tab)
+{
+  g_return_if_fail (VINAGRE_IS_TAB (tab));
+
+  return tab->priv->state;
+}
+
+VinagreTab *
+vinagre_tab_get_from_connection (VinagreConnection *conn)
+{
+  gpointer res;
+	
+  g_return_val_if_fail (VINAGRE_IS_CONNECTION (conn), NULL);
+	
+  res = g_object_get_data (G_OBJECT (conn), VINAGRE_TAB_KEY);
+	
+  return (res != NULL) ? VINAGRE_TAB (res) : NULL;
+}
+
+gboolean
+vinagre_tab_is_pointer_grab (VinagreTab *tab) {
+  g_return_if_fail (VINAGRE_IS_TAB (tab));
+
+  return tab->priv->pointer_grab;
+}
+
+/* vim: set ts=8: */
