@@ -32,10 +32,11 @@
 #include "vinagre-window.h"
 #include "vinagre-notebook.h"
 #include "vinagre-fav.h"
-#include "vinagre-prefs-manager.h"
+#include "vinagre-prefs.h"
 #include "vinagre-utils.h"
 #include "vinagre-bookmarks.h"
 #include "vinagre-ui.h"
+#include "vinagre-mdns.h"
 
 #include "vinagre-window-private.h"
 
@@ -46,13 +47,13 @@
 G_DEFINE_TYPE(VinagreWindow, vinagre_window, GTK_TYPE_WINDOW)
 
 static void
-vinagre_window_finalize (GObject *object)
+vinagre_window_dispose (GObject *object)
 {
   VinagreWindow *window = VINAGRE_WINDOW (object);
 
   if (window->priv->fav_conn_selected)
     {
-      vinagre_connection_free (window->priv->fav_conn_selected);
+      g_object_unref (window->priv->fav_conn_selected);
       window->priv->fav_conn_selected = NULL;
     }
 
@@ -62,9 +63,37 @@ vinagre_window_finalize (GObject *object)
       window->priv->manager = NULL;
     }
 
+  if (window->priv->signal_notebook != 0)
+    {
+      g_signal_handler_disconnect (window->priv->notebook,
+				   window->priv->signal_notebook);
+      window->priv->signal_notebook = 0;
+    }
 
-  vinagre_bookmarks_finalize ();
+  if (window->priv->signal_clipboard != 0)
+    {
+      GtkClipboard  *cb = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);;
 
+      g_signal_handler_disconnect (cb,
+				   window->priv->signal_clipboard);
+      window->priv->signal_clipboard = 0;
+    }
+
+  if (window->priv->update_recents_menu_ui_id != 0)
+    {
+      GtkRecentManager *recent_manager = gtk_recent_manager_get_default ();
+
+      g_signal_handler_disconnect (recent_manager,
+				   window->priv->update_recents_menu_ui_id);
+      window->priv->update_recents_menu_ui_id = 0;
+    }
+
+  G_OBJECT_CLASS (vinagre_window_parent_class)->dispose (object);
+}
+
+static void
+vinagre_window_finalize (GObject *object)
+{
   G_OBJECT_CLASS (vinagre_window_parent_class)->finalize (object);
 }
 
@@ -73,20 +102,8 @@ vinagre_window_delete_event (GtkWidget   *widget,
 			     GdkEventAny *event)
 {
   VinagreWindow *window = VINAGRE_WINDOW (widget);
-  GtkClipboard  *cb;
-
-  cb = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
-
-  if (window->priv->signal_notebook > 0)
-    g_signal_handler_disconnect (window->priv->notebook,
-				 window->priv->signal_notebook);
-
-  if (window->priv->signal_clipboard > 0)
-    g_signal_handler_disconnect (cb,
-				 window->priv->signal_clipboard);
 
   vinagre_window_close_all_tabs (window);
-  gtk_main_quit ();
 
   if (GTK_WIDGET_CLASS (vinagre_window_parent_class)->delete_event)
     return GTK_WIDGET_CLASS (vinagre_window_parent_class)->delete_event (widget, event);
@@ -126,7 +143,7 @@ vinagre_window_show_hide_controls (VinagreWindow *window)
       if (window->priv->statusbar_visible)
         gtk_widget_show_all (window->priv->statusbar);
 
-      gtk_notebook_set_show_tabs (GTK_NOTEBOOK (window->priv->notebook), TRUE);
+      vinagre_notebook_show_hide_tabs (VINAGRE_NOTEBOOK (window->priv->notebook));
       gtk_notebook_set_show_border (GTK_NOTEBOOK (window->priv->notebook), TRUE);
     }
 }
@@ -135,21 +152,21 @@ static gboolean
 vinagre_window_state_event_cb (GtkWidget *widget,
 			       GdkEventWindowState *event)
 {
-  VinagreWindow *window;
+  VinagreWindow *window = VINAGRE_WINDOW (widget);
+
+  window->priv->window_state = event->new_window_state;
+  g_object_set (vinagre_prefs_get_default (),
+		"window-state", window->priv->window_state,
+		NULL);
 
   if ((event->changed_mask & GDK_WINDOW_STATE_FULLSCREEN) == 0)
     return FALSE;
 
-  window = VINAGRE_WINDOW (widget);
-
   if (event->new_window_state & GDK_WINDOW_STATE_FULLSCREEN)
     window->priv->fullscreen = TRUE;
   else
     window->priv->fullscreen = FALSE;
 
-  window->priv->window_state = event->new_window_state;
-  vinagre_prefs_manager_set_window_state (window->priv->window_state);
-
   vinagre_window_show_hide_controls (window);
 
   if (GTK_WIDGET_CLASS (vinagre_window_parent_class)->window_state_event)
@@ -167,39 +184,23 @@ vinagre_window_configure_event (GtkWidget         *widget,
   window->priv->width  = event->width;
   window->priv->height = event->height;
 
-  vinagre_prefs_manager_set_window_size (window->priv->width,
-					 window->priv->height);
+  g_object_set (vinagre_prefs_get_default (),
+		"window-width", window->priv->width,
+		"window-height", window->priv->height,
+		NULL);
 
   return GTK_WIDGET_CLASS (vinagre_window_parent_class)->configure_event (widget, event);
 }
 
-
-static gboolean
-vinagre_window_key_press_cb (GtkWidget   *widget,
-			     GdkEventKey *event)
-{
-  VinagreWindow *window = VINAGRE_WINDOW (widget);
-
-  switch (event->keyval)
-    {
-      case GDK_F11:
-	if (window->priv->active_tab)
-	  vinagre_window_toggle_fullscreen (window);
-	return FALSE;
-    }
-
-  return GTK_WIDGET_CLASS (vinagre_window_parent_class)->key_press_event (widget, event);
-}
-
 static void
 vinagre_window_class_init (VinagreWindowClass *klass)
 {
   GObjectClass *object_class    = G_OBJECT_CLASS (klass);
   GtkWidgetClass *widget_class  = GTK_WIDGET_CLASS (klass);
 
-  object_class->finalize     = vinagre_window_finalize;
+  object_class->finalize = vinagre_window_finalize;
+  object_class->dispose  = vinagre_window_dispose;
 
-  widget_class->key_press_event    = vinagre_window_key_press_cb;
   widget_class->window_state_event = vinagre_window_state_event_cb;
   widget_class->configure_event    = vinagre_window_configure_event;
   widget_class->delete_event       = vinagre_window_delete_event;
@@ -268,14 +269,30 @@ static void
 activate_recent_cb (GtkRecentChooser *action, VinagreWindow *window)
 {
   VinagreConnection *conn;
+  gchar             *error, *msg;
 
-  conn = vinagre_connection_new_from_string (gtk_recent_chooser_get_current_uri (action));
-  vinagre_cmd_open_bookmark (window, conn);
+  conn = vinagre_connection_new_from_string (gtk_recent_chooser_get_current_uri (action),
+					     &error);
+  if (conn)
+    {
+      vinagre_cmd_open_bookmark (window, conn);
+      g_object_unref (conn);
+    }
+  else
+    {
+      msg = g_strdup_printf ("%s %s",
+			     _("The following error has occurred:"),
+			     error ? error : _("Unknown error"));
+      vinagre_utils_show_error (msg, GTK_WINDOW (window));
+      g_free (msg);
+    }
 
-  vinagre_connection_free (conn);
+  if (error)
+    g_free (error);
 }
 
-static void update_recent_connections (VinagreWindow *window)
+static void
+update_recent_connections (VinagreWindow *window)
 {
   VinagreWindowPrivate *p = window->priv;
 
@@ -302,6 +319,19 @@ recent_manager_changed (GtkRecentManager *manager,
   update_recent_connections (window);
 }
 
+static void
+show_hide_accels (VinagreWindow *window)
+{
+  gboolean show_accels;
+
+  g_object_get (vinagre_prefs_get_default (),
+		"show-accels", &show_accels,
+		NULL);
+  g_object_set (gtk_settings_get_default (),
+		"gtk-enable-accels", show_accels,
+		NULL);
+}
+
 static void
 create_menu_bar_and_toolbar (VinagreWindow *window, 
 			     GtkWidget     *main_box)
@@ -369,6 +399,10 @@ create_menu_bar_and_toolbar (VinagreWindow *window,
 				vinagre_machine_connected_menu_entries,
 				G_N_ELEMENTS (vinagre_machine_connected_menu_entries),
 				window);
+  gtk_action_group_add_toggle_actions  (action_group,
+					vinagre_machine_connected_toggle_menu_entries,
+					G_N_ELEMENTS (vinagre_machine_connected_toggle_menu_entries),
+					window);
 
   gtk_ui_manager_insert_action_group (manager, action_group, 0);
   g_object_unref (action_group);
@@ -378,6 +412,8 @@ create_menu_bar_and_toolbar (VinagreWindow *window,
   g_object_set (action, "is_important", TRUE, NULL);
   action = gtk_action_group_get_action (action_group, "MachineTakeScreenshot");
   g_object_set (action, "is_important", TRUE, NULL);
+  action = gtk_action_group_get_action (action_group, "MachineSendCtrlAltDel");
+  g_object_set (action, "is_important", TRUE, NULL);
 
   /* now load the UI definition */
   gtk_ui_manager_add_ui_from_file (manager, vinagre_utils_get_ui_xml_filename (), &error);
@@ -413,6 +449,14 @@ create_menu_bar_and_toolbar (VinagreWindow *window,
   window->priv->recent_action = gtk_recent_action_new ("recent_connections",
 						     _("_Recent connections"),
 						       NULL, NULL);
+  g_object_set (G_OBJECT (window->priv->recent_action),
+		"show-not-found", TRUE,
+		"local-only", FALSE,
+		"show-private", TRUE,
+		"show-tips", TRUE,
+		"sort-type", GTK_RECENT_SORT_MRU,
+		NULL);
+  gtk_recent_chooser_set_local_only (GTK_RECENT_CHOOSER (window->priv->recent_action), FALSE);
   g_signal_connect (window->priv->recent_action,
 		    "item-activated",
 		    G_CALLBACK (activate_recent_cb),
@@ -435,15 +479,21 @@ create_menu_bar_and_toolbar (VinagreWindow *window,
   update_recent_connections (window);
 
   recent_manager = gtk_recent_manager_get_default ();
-  g_signal_connect (recent_manager,
+  window->priv->update_recents_menu_ui_id = g_signal_connect (
+		    recent_manager,
 		    "changed",
 		    G_CALLBACK (recent_manager_changed),
 		    window);
 
+  g_signal_connect_swapped (vinagre_prefs_get_default (),
+			    "notify::show-accels",
+			     G_CALLBACK (show_hide_accels),
+			     window);
+  show_hide_accels (window);
 }
 
-static void
-set_machine_menu_sensitivity (VinagreWindow *window)
+void
+vinagre_window_update_machine_menu_sensitivity (VinagreWindow *window)
 {
   gboolean active;
 
@@ -452,7 +502,8 @@ set_machine_menu_sensitivity (VinagreWindow *window)
   active = gtk_notebook_get_n_pages (GTK_NOTEBOOK (window->priv->notebook)) > 0;
   gtk_action_group_set_sensitive (window->priv->action_group, active);
 
-  active = window->priv->machines_connected > 0;
+  active = (window->priv->active_tab) &&
+	   (vinagre_tab_get_state (VINAGRE_TAB (window->priv->active_tab)) == VINAGRE_TAB_STATE_CONNECTED);
   gtk_action_group_set_sensitive (window->priv->machine_connected_action_group, active);
 }
 
@@ -463,7 +514,9 @@ fav_panel_size_allocate (GtkWidget     *widget,
 {
   window->priv->side_panel_size = allocation->width;
   if (window->priv->side_panel_size > 0)
-    vinagre_prefs_manager_set_side_panel_size (window->priv->side_panel_size);
+    g_object_set (vinagre_prefs_get_default (),
+		  "side-panel-size", window->priv->side_panel_size,
+		  NULL);
 }
 
 static void
@@ -471,9 +524,6 @@ fav_panel_activated (VinagreFav        *fav,
 		     VinagreConnection *conn,
 		     VinagreWindow     *window)
 {
-  g_return_if_fail (VINAGRE_IS_FAV (fav));
-  g_return_if_fail (VINAGRE_IS_WINDOW (window));
-
   vinagre_cmd_open_bookmark (window, conn);
 }
 
@@ -491,7 +541,7 @@ fav_panel_selected (VinagreFav        *fav,
 
   if (window->priv->fav_conn_selected)
     {
-      vinagre_connection_free (window->priv->fav_conn_selected);
+      g_object_unref (window->priv->fav_conn_selected);
       window->priv->fav_conn_selected = NULL;
     }
 
@@ -514,13 +564,10 @@ void
 vinagre_window_update_bookmarks_list_menu (VinagreWindow *window)
 {
   VinagreWindowPrivate *p = window->priv;
-  GList                *actions, *l, *favs;
-  gint                  n, i;
-  guint                 id;
-  VinagreConnection    *conn;
-  gchar                *action_name, *action_label;
-  GtkAction            *action;
-  gchar                *name, *tooltip;
+  GList  *actions, *l;
+  GSList *favs, *mdnss;
+  gint   n, m, i;
+  guint  id;
 
   g_return_if_fail (p->bookmarks_list_action_group != NULL);
 
@@ -536,38 +583,42 @@ vinagre_window_update_bookmarks_list_menu (VinagreWindow *window)
       gtk_action_group_remove_action (p->bookmarks_list_action_group,
 				      GTK_ACTION (l->data));
     }
-    g_list_free (actions);
+  g_list_free (actions);
 
-  /* Get the bookmarks from file */
-  favs = vinagre_bookmarks_get_all ();
-  n = g_list_length (favs);
+  favs = vinagre_bookmarks_get_all (vinagre_bookmarks_get_default ());
+  mdnss = vinagre_mdns_get_all (vinagre_mdns_get_default ());
+  n = g_slist_length (favs);
+  m = g_slist_length (mdnss);
   i = 0;
 
-  id = (n > 0) ? gtk_ui_manager_new_merge_id (p->manager) : 0;
+  id = (n > 0||m > 0) ? gtk_ui_manager_new_merge_id (p->manager) : 0;
 
   while (favs)
     {
-      conn = (VinagreConnection *) favs->data;
-      name = vinagre_connection_best_name (conn);
+      VinagreConnection *conn;
+      gchar             *action_name, *action_label;
+      GtkAction         *action;
+      gchar             *name, *tooltip;
 
-      action_name = g_strdup_printf ("Fav_%d", i);
+      conn = (VinagreConnection *) favs->data;
+      g_assert (VINAGRE_IS_CONNECTION (conn));
+
+      name = vinagre_connection_get_best_name (conn);
+
+      action_name = g_strdup_printf ("Bookmark_%d", i);
       /* Translators: This is server:port, a statusbar tooltip when mouse is over a bookmark item on menu */
-      tooltip = g_strdup_printf (_("Open %s:%d"), conn->host, conn->port);
-      action_label = vinagre_utils_escape_underscores (
-		     name,
-		     -1);
+      tooltip = g_strdup_printf (_("Open %s:%d"),
+                                 vinagre_connection_get_host (conn),
+                                 vinagre_connection_get_port (conn));
+      action_label = vinagre_utils_escape_underscores (name, -1);
       action = gtk_action_new (action_name,
 			       action_label,
 			       tooltip,
 			       NULL);
       g_object_set (G_OBJECT (action), "icon-name", "application-x-vnc", NULL);
-      g_object_set_data_full (G_OBJECT (action),
-			      "conn",
-			      conn,
-			      (GDestroyNotify) vinagre_connection_free);
+      g_object_set_data (G_OBJECT (action), "conn", conn);
       gtk_action_group_add_action (p->bookmarks_list_action_group,
 				   GTK_ACTION (action));
-
       g_signal_connect (action,
 			"activate",
 			G_CALLBACK (vinagre_cmd_bookmarks_open),
@@ -590,7 +641,55 @@ vinagre_window_update_bookmarks_list_menu (VinagreWindow *window)
       i++;
     }
 
-  g_list_free (favs);
+  /* avahi */
+  i = 0;
+  while (mdnss)
+    {
+      VinagreConnection *conn;
+      gchar             *action_name, *action_label;
+      GtkAction         *action;
+      gchar             *name, *tooltip;
+
+      conn = (VinagreConnection *) mdnss->data;
+      g_assert (VINAGRE_IS_CONNECTION (conn));
+
+      name = vinagre_connection_get_best_name (conn);
+
+      action_name = g_strdup_printf ("Avahi_%d", i);
+      /* Translators: This is server:port, a statusbar tooltip when mouse is over a bookmark item on menu */
+        tooltip = g_strdup_printf (_("Open %s:%d"),
+                                 vinagre_connection_get_host (conn),
+                                 vinagre_connection_get_port (conn));
+      action_label = vinagre_utils_escape_underscores (name, -1);
+      action = gtk_action_new (action_name,
+			       action_label,
+			       tooltip,
+			       NULL);
+      g_object_set (G_OBJECT (action), "icon-name", "application-x-vnc", NULL);
+      g_object_set_data (G_OBJECT (action), "conn", conn);
+      gtk_action_group_add_action (p->bookmarks_list_action_group,
+				   GTK_ACTION (action));
+      g_signal_connect (action,
+			"activate",
+			G_CALLBACK (vinagre_cmd_bookmarks_open),
+			window);
+
+      gtk_ui_manager_add_ui (p->manager,
+			     id,
+			     "/MenuBar/BookmarksMenu/AvahiList",
+			     action_name, action_name,
+			     GTK_UI_MANAGER_MENUITEM,
+			     FALSE);
+
+      g_object_unref (action);
+      g_free (action_name);
+      g_free (action_label);
+      g_free (name);
+      g_free (tooltip);
+
+      mdnss = mdnss->next;
+      i++;
+    }
 
   p->bookmarks_list_menu_ui_id = id;
 }
@@ -605,7 +704,9 @@ create_side_panel (VinagreWindow *window)
 		   FALSE, 
 		   FALSE);
 
-  window->priv->side_panel_size = vinagre_prefs_manager_get_side_panel_size ();
+  g_object_get (vinagre_prefs_get_default (),
+		"side-panel-size", &(window->priv->side_panel_size),
+		NULL);
   gtk_paned_set_position (GTK_PANED (window->priv->hpaned), window->priv->side_panel_size);
 
   g_signal_connect (window->priv->fav_panel,
@@ -629,6 +730,7 @@ init_widgets_visibility (VinagreWindow *window)
   gint w, h;
   GtkAction *action;
   gboolean visible;
+  VinagrePrefs *prefs = vinagre_prefs_get_default ();
 
   /* Remove and Edit bookmarks starts disabled */
   action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
@@ -639,64 +741,32 @@ init_widgets_visibility (VinagreWindow *window)
 					"BookmarksEdit");
   gtk_action_set_sensitive (action, FALSE);
 
-  /* fav panel visibility */
+  /* side panel visibility */
   action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
-					"ViewBookmarks");
-  visible = vinagre_prefs_manager_get_side_pane_visible ();
+					"ViewSidePanel");
+  g_object_get (prefs, "side-panel-visible", &visible, NULL);
   if (gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)) != visible)
     gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), visible);
 
   /* toolbar visibility */
   action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
 					"ViewToolbar");
-  visible = vinagre_prefs_manager_get_toolbar_visible ();
+  g_object_get (prefs, "toolbar-visible", &visible, NULL);
   if (gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)) != visible)
     gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), visible);
 
   /* statusbar visibility */
   action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
 					"ViewStatusbar");
-  visible = vinagre_prefs_manager_get_side_pane_visible ();
+  g_object_get (prefs, "statusbar-visible", &visible, NULL);
   if (gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)) != visible)
     gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), visible);
 
-  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
-				vinagre_prefs_manager_get_statusbar_visible ());
-
-//  if (vinagre_prefs_manager_get_side_pane_visible ())
-//    gtk_widget_show_all (window->priv->fav_panel);
-//  else
-//    {
-//      gtk_widget_hide (window->priv->fav_panel);
-//printf ("hiding... was visible: %d\n", GTK_WIDGET_VISIBLE(window->priv->fav_panel));
-
-//      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
-//    }
-
-  /* toolbar visibility */
-//  action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
-//					"ViewToolbar");
-//  if (vinagre_prefs_manager_get_toolbar_visible ())
-//    gtk_widget_show_all (window->priv->toolbar);
-//  else
-//    {
-//      gtk_widget_hide (window->priv->toolbar);
-//      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
-//   }
-
-  /* toolbar visibility */
-//  action = gtk_action_group_get_action (window->priv->always_sensitive_action_group,
-//					"ViewStatusbar");
-//  if (vinagre_prefs_manager_get_statusbar_visible ())
-//    gtk_widget_show_all (window->priv->statusbar);
-//  else
-//    {
- //     gtk_widget_hide (window->priv->statusbar);
-//      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
- //   }
-
-  state = vinagre_prefs_manager_get_window_state ();
-  vinagre_prefs_manager_get_window_size (&w, &h);
+  g_object_get (prefs,
+		"window-state", &state,
+		"window-width", &w,
+		"window-height", &h,
+		NULL);
 
   gtk_window_set_default_size (GTK_WINDOW (window), w, h);
 
@@ -727,7 +797,7 @@ create_statusbar (VinagreWindow *window,
 void
 vinagre_window_set_title (VinagreWindow *window)
 {
-  gchar *title, *name;
+  gchar *title, *name, *grab;
 
   if (window->priv->active_tab == NULL)
     {
@@ -735,13 +805,48 @@ vinagre_window_set_title (VinagreWindow *window)
       return;
     }
 
-  name = vinagre_connection_best_name (vinagre_tab_get_conn (VINAGRE_TAB (window->priv->active_tab)));
-  title = g_strdup_printf ("%s - %s",
+  if (vinagre_tab_is_pointer_grab (VINAGRE_TAB (window->priv->active_tab)))
+    grab = g_strdup_printf (" (%s)", _("Press Ctrl+Alt to release the cursor"));
+  else
+    grab = g_strdup ("");
+
+  name = vinagre_connection_get_best_name (vinagre_tab_get_conn (VINAGRE_TAB (window->priv->active_tab)));
+  title = g_strdup_printf ("%s%s - %s",
 			   name,
+			   grab,
 			   g_get_application_name ());
   gtk_window_set_title (GTK_WINDOW (window), title);
   g_free (title);
   g_free (name);
+  g_free (grab);
+}
+
+static void
+update_toggle_machine_items (VinagreWindow *window)
+{
+  GtkAction *action;
+
+  g_return_if_fail (VINAGRE_IS_WINDOW (window));
+
+  if (window->priv->active_tab == NULL)
+    {
+      action = gtk_action_group_get_action (window->priv->machine_connected_action_group, "ViewScaling");
+      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
+
+      action = gtk_action_group_get_action (window->priv->machine_connected_action_group, "ViewReadOnly");
+      gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
+
+      return;
+    }
+
+  action = gtk_action_group_get_action (window->priv->machine_connected_action_group, "ViewScaling");
+  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
+				vinagre_tab_get_scaling (VINAGRE_TAB (window->priv->active_tab)));
+
+  action = gtk_action_group_get_action (window->priv->machine_connected_action_group, "ViewReadOnly");
+  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
+				vinagre_tab_get_readonly (VINAGRE_TAB (window->priv->active_tab)));
+
 }
 
 static void
@@ -750,15 +855,21 @@ vinagre_window_page_removed (GtkNotebook   *notebook,
 			     guint         page_num,
 			     VinagreWindow *window)
 {
+  GtkNotebook *nb;
+
   g_return_if_fail (VINAGRE_IS_WINDOW (window));
 
-  window->priv->active_tab = gtk_notebook_get_nth_page (GTK_NOTEBOOK (window->priv->notebook), gtk_notebook_get_current_page (GTK_NOTEBOOK (window->priv->notebook)));
+  nb = GTK_NOTEBOOK (window->priv->notebook);
+
+  window->priv->active_tab = gtk_notebook_get_nth_page (nb,
+							gtk_notebook_get_current_page (nb));
 
   if (!window->priv->active_tab && window->priv->fullscreen)
     vinagre_window_toggle_fullscreen (window);
 
   vinagre_window_set_title (window);
-  set_machine_menu_sensitivity (window);
+  update_toggle_machine_items (window);
+  vinagre_window_update_machine_menu_sensitivity (window);
 }
 
 static void
@@ -772,7 +883,8 @@ vinagre_window_page_added (GtkNotebook  *notebook,
   window->priv->active_tab = child;
 
   vinagre_window_set_title (window);
-  set_machine_menu_sensitivity (window);
+  update_toggle_machine_items (window);
+  vinagre_window_update_machine_menu_sensitivity (window);
 }
 
 static void 
@@ -792,6 +904,8 @@ vinagre_window_switch_page (GtkNotebook     *notebook,
   window->priv->active_tab = tab;
 
   vinagre_window_set_title (window);
+  update_toggle_machine_items (window);
+  vinagre_window_update_machine_menu_sensitivity (window);
 }
 
 static void
@@ -828,13 +942,18 @@ vinagre_window_clipboard_cb (GtkClipboard *cb, GdkEvent *event, VinagreWindow *w
 {
   gchar *text;
 
-  text = gtk_clipboard_wait_for_text (cb);
-  if (text) {
-    if (window->priv->active_tab)
-      vinagre_tab_paste_text (VINAGRE_TAB (window->priv->active_tab), text);
-    g_free (text);
-  }
+  if (!window->priv->active_tab)
+    return;
 
+  if (gtk_clipboard_get_owner (cb) == G_OBJECT (window->priv->active_tab))
+    return;
+
+  text = gtk_clipboard_wait_for_text (cb);
+  if (!text)
+    return;
+
+  vinagre_tab_paste_text (VINAGRE_TAB (window->priv->active_tab), text);
+  g_free (text);
 }
 
 static void
@@ -861,9 +980,6 @@ vinagre_window_init (VinagreWindow *window)
   window->priv->fav_conn_selected = NULL;
   window->priv->fullscreen = FALSE;
   window->priv->signal_notebook = 0;
-  window->priv->machines_connected = 0;
-
-  vinagre_bookmarks_init ();
 
   main_box = gtk_vbox_new (FALSE, 0);
   gtk_container_add (GTK_CONTAINER (window), main_box);
@@ -894,9 +1010,18 @@ vinagre_window_init (VinagreWindow *window)
   gtk_widget_grab_focus (window->priv->hpaned);
 
   init_widgets_visibility (window);
-  set_machine_menu_sensitivity (window);
+  vinagre_window_update_machine_menu_sensitivity (window);
 
   vinagre_window_update_bookmarks_list_menu (window);
+  g_signal_connect_swapped (vinagre_bookmarks_get_default (),
+                            "changed",
+                            G_CALLBACK (vinagre_window_update_bookmarks_list_menu),
+                            window);
+  g_signal_connect_swapped (vinagre_mdns_get_default (),
+                            "changed",
+                            G_CALLBACK (vinagre_window_update_bookmarks_list_menu),
+                            window);
+
   vinagre_window_init_clipboard (window);
 }
 
@@ -1009,6 +1134,14 @@ vinagre_window_get_sensitive_action (VinagreWindow *window)
   return window->priv->always_sensitive_action_group;
 }
 
+GtkActionGroup *
+vinagre_window_get_connected_action (VinagreWindow *window)
+{
+  g_return_val_if_fail (VINAGRE_IS_WINDOW (window), NULL);
+
+  return window->priv->machine_connected_action_group;
+}
+
 GtkWidget *
 vinagre_window_get_fav_panel (VinagreWindow *window)
 {
@@ -1027,7 +1160,7 @@ vinagre_window_close_active_tab	(VinagreWindow *window)
 }
 
 VinagreTab *
-vinagre_window_get_active_tab	(VinagreWindow *window)
+vinagre_window_get_active_tab (VinagreWindow *window)
 {
   g_return_val_if_fail (VINAGRE_IS_WINDOW (window), NULL);
 
@@ -1042,25 +1175,60 @@ vinagre_window_get_ui_manager (VinagreWindow *window)
   return window->priv->manager;
 }
 
-void
-_vinagre_window_add_machine_connected (VinagreWindow *window)
+static void
+add_connection (VinagreTab *tab, GList **res)
 {
-  g_return_if_fail (VINAGRE_IS_WINDOW (window));
-
-  window->priv->machines_connected++;
-  set_machine_menu_sensitivity (window);
+  VinagreConnection *conn;
+	
+  conn = vinagre_tab_get_conn (tab);
+	
+  *res = g_list_prepend (*res, conn);
 }
 
-void
-_vinagre_window_del_machine_connected (VinagreWindow *window)
+/* Returns a newly allocated list with all the connections in the window */
+GList *
+vinagre_window_get_connections (VinagreWindow *window)
 {
-  g_return_if_fail (VINAGRE_IS_WINDOW (window));
+  GList *res = NULL;
 
-  if (window->priv->machines_connected > 0)
+  g_return_val_if_fail (VINAGRE_IS_WINDOW (window), NULL);
+	
+  gtk_container_foreach (GTK_CONTAINER (window->priv->notebook),
+			 (GtkCallback)add_connection,
+			  &res);
+			       
+  res = g_list_reverse (res);
+	
+  return res;
+}
+
+VinagreTab *
+vinagre_window_conn_exists (VinagreWindow *window, VinagreConnection *conn)
+{
+  VinagreConnection *c;
+  VinagreTab *tab = NULL;
+  GList *l;
+
+  g_return_val_if_fail (VINAGRE_IS_WINDOW (window), NULL);
+  g_return_val_if_fail (VINAGRE_IS_CONNECTION (conn), NULL);
+
+  l = vinagre_window_get_connections (window);
+
+  while (l != NULL)
     {
-      window->priv->machines_connected--;
-      set_machine_menu_sensitivity (window);
+      c = VINAGRE_CONNECTION (l->data);
+
+      if (!strcmp (vinagre_connection_get_host (conn), vinagre_connection_get_host (c)) &&
+	  vinagre_connection_get_port (conn) == vinagre_connection_get_port (c))
+	{
+	  tab = g_object_get_data (G_OBJECT (c), VINAGRE_TAB_KEY);
+	  break;
+	}
+      l = l->next;
     }
+  g_list_free (l);
+
+  return tab;
 }
 
-/* vim: ts=8 */
+/* vim: set ts=8: */
